/* Simple monitor for the new twtch stuff */

/*
 * Copyright (c) 2000  En Garde Systems. All rights reserved.
 *    
 * Redistribution and use in source and binary forms, with or without    
 * modification, for NON COMMERCIAL USE are permitted provided that: 
 * (1) source code distributions retain the above copyright notice and this 
 * paragraph in its entirety, and (2) distributions including binary code 
 * include the above copyright notice and this paragraph in its entirety in 
 * the documentation or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * For commercial use or modification of this program, contact:
 * En Garde Systems
 * 4848 Tramway Ridge, Suite 122
 * Albuquerque, NM 87111
 */
 
/* 
   K2,
   Some quick hack to let us log what's going on inside pty's
   Version: 0.3.6
   remember to change the hostname!!!
*/
#define LOGDIR "/var/adm/doord/"
#define SIZE 2048
#define NPTY 64
#define PORT 54242

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#define _REENTRANT
#include <dirent.h>
#include <thread.h>
#include <procfs.h>
#include <unistd.h>
#include <stropts.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/stream.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <netdb.h>
#include <signal.h>

#include "twtch.h"


void *allocate();

thread_t tid;
int fd,fdX;

struct ptys {
   int inuse[NPTY];
   int uid[NPTY];
   int euid[NPTY];
} ptys;

int main()
{
   int flags,n,s,len;
   char zbuf[SIZE+4];
   char buffer[SIZE+1];
   struct strbuf dataptr;
   char hostname[64] = "loghost";
   extern int errno;
   struct packet *sp;
   struct hostent *hp;
   struct sockaddr_in name;
#ifdef LOCAL
   FILE *logfile;
   char fsys[] = "-FROM_SYS";
   char fuser[] = "-FROM_USER";
#endif


   sigignore(SIGTTOU);
   
   /* this will manage all our pty's */
   thr_create(NULL, 0, allocate, 0, 0,&tid);

   if ((fd=open("/dev/doord", O_RDONLY))<0) {
   	perror("Can't open /dev/doord");
   	exit(1);
   }
   dataptr.maxlen=SIZE-1;
   dataptr.buf=buffer;
   flags=0;   
#ifndef LOCAL
   if((hp = gethostbyname(hostname)) == NULL) {
      perror("unknown hostname");
      exit(1);
   }

   memset(&name,0,sizeof(struct sockaddr_in));   
   name.sin_family = AF_INET;
   name.sin_port = htons(PORT);
   memcpy(&name.sin_addr, hp->h_addr_list[0],hp->h_length);
   len = sizeof(struct sockaddr_in);
#endif    

   for (;;) {
	while (getmsg(fd, NULL, &dataptr, &flags)>=0) {
		sp=(struct packet *)dataptr.buf;
#ifdef LOCAL
     if(sp->from == FROM_SYS)
         snprintf(zbuf,1023,"%s%d%s:%d:%d",LOGDIR,(sp->dev & 0xff),fsys,ptys.uid[(sp->dev & 0xff)],ptys.euid[(sp->dev & 0xff)]);
      else /* FROM_USER ;) */
         snprintf(zbuf,1023,"%s%d%s:%d:%d",LOGDIR,(sp->dev & 0xff),fuser,ptys.uid[(sp->dev & 0xff)],ptys.euid[(sp->dev & 0xff)]);
         
      logfile = fopen(zbuf,"ab+"); 
      fprintf(logfile,"%s",dataptr.buf+sizeof(struct packet)); fclose(logfile);
#else      
      if(sp->from == FROM_SYS)
         n = snprintf(zbuf,SIZE+3,"%s%.5d%.5d%.2d%s","S",ptys.uid[(sp->dev & 0xff)],ptys.euid[(sp->dev & 0xff)],(sp->dev & 0xff),dataptr.buf+sizeof(struct packet));
      else /* FROM_USER ;) */
         n = snprintf(zbuf,SIZE+3,"%s%.5d%.5d%.2d%s","U",ptys.uid[(sp->dev & 0xff)],ptys.euid[(sp->dev & 0xff)],(sp->dev & 0xff),dataptr.buf+sizeof(struct packet));
      if((s = socket(AF_INET, SOCK_STREAM,0)) < 0) {
         perror("socket");
         exit(1);
      }      
      if(connect(s, (struct sockaddr *) &name, len) < 0) {
         perror("connect");
         exit(1);
      }
      if(send(s,zbuf,n,0) < 0) {
         perror("send");
         exit(1);
      }      
      close(s);
      memset(zbuf,0,n);
#endif         
      memset(dataptr.buf,0,dataptr.len); 
	}
   
	if (errno!=EWOULDBLOCK) {
		perror("getmsg");
		exit(1);
	} else
		sleep(1);
   }
   exit(0);
}

/*  
   OK, here's the rewrite.  find allocated pty's from proc.
   allocate and deallocate on the fly. maintain list of inuse pty's.
*/
void *allocate()
{
   psinfo_t ps;
   int procfd,i,stillinuse;
   char name[128];
   char port[128];
   char procdir[] = "/proc/";
   char devdir[] = "/dev/";
   DIR *dirp;
   struct dirent *dentp;

/* infinate loop */
while(1) {
   dirp = opendir(procdir);
   while (dentp = readdir(dirp)) {
      if (dentp->d_name[0] == '.')
			continue;
      
      snprintf(name,127,"%s%s%s",procdir,dentp->d_name,"/psinfo");
      procfd = open(name,O_RDONLY);
      read(procfd, (char *)&ps, sizeof(ps));
      if((ps.pr_ttydev != PRNODEV) && (ps.pr_ttydev != 0)) {
         snprintf(port,127,"%spts/%d",devdir,(ps.pr_ttydev & 0xff));

         ptys.inuse[(ps.pr_ttydev & 0xff)] = 1;
         ptys.uid[(ps.pr_ttydev & 0xff)] = ps.pr_uid;
         ptys.euid[(ps.pr_ttydev & 0xff)] = ps.pr_euid;
         
         fdX=open(port, O_WRONLY);
         ioctl(fdX, I_LOOK, "doord");
         ioctl(fdX, I_PUSH, "doord");
#ifdef DEBUG
         fprintf(stderr,"PUSH'd %s\n",port);
#endif
         close(fdX);
      }
      close(procfd);
   }
   closedir(dirp);
   
   sleep(2);

   /* cycle through all possiable PTY's looking for inuse ones, check if still active */
   for(i=0;i<NPTY;i++) {
      stillinuse=0;
      if(ptys.inuse[i] == 1) {
         dirp = opendir(procdir);
         while (dentp = readdir(dirp)) {         
            if (dentp->d_name[0] == '.')
   		   	continue;
      
            snprintf(name,127,"%s%s%s",procdir,dentp->d_name,"/psinfo");
            procfd = open(name,O_RDONLY);
            read(procfd, (char *)&ps, sizeof(ps));
            if((ps.pr_ttydev & 0xff) == i) {
               stillinuse=1;
               }
            close(procfd);
         }
         closedir(dirp);
         
         /* if not detected active, deactiveate */
         if(stillinuse == 0) {
            snprintf(port,127,"%spts/%d",devdir,i);
            fdX=open(port, O_WRONLY|O_NDELAY);
            ioctl(fdX, I_LOOK, "doord");
            ioctl(fdX, I_POP, 0);
            ptys.inuse[i] = 0;
#ifdef DEBUG
            fprintf(stderr,"POP'd %s\n",port);
#endif
            close(fdX);
         }
      }
   }
}
/* end of infinate loop hehe */
}
